@@ -1,7 +1,7 @@
 /*
  * SPDX-License-Identifier: ISC
  *
- * Copyright (c) 2019-2021 Todd C. Miller <Todd.Miller@sudo.ws>
+ * Copyright (c) 2019-2022 Todd C. Miller <Todd.Miller@sudo.ws>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -16,7 +16,12 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-#include "config.h"
+/*
+ * This is an open source non-commercial project. Dear PVS-Studio, please check it.
+ * PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com
+ */
+
+#include <config.h>
 
 #include <sys/stat.h>
 #include <sys/types.h>
@@ -30,7 +35,7 @@
 #ifdef HAVE_STDBOOL_H
 # include <stdbool.h>
 #else
-# include "compat/stdbool.h"
+# include <compat/stdbool.h>
 #endif /* HAVE_STDBOOL_H */
 #include <stdio.h>
 #include <stdlib.h>
@@ -38,33 +43,39 @@
 #include <time.h>
 #include <unistd.h>
 
-#include "sudo_compat.h"
-#include "sudo_debug.h"
-#include "sudo_eventlog.h"
-#include "sudo_gettext.h"
-#include "sudo_iolog.h"
-#include "sudo_fatal.h"
-#include "sudo_queue.h"
-#include "sudo_util.h"
+#include <sudo_compat.h>
+#include <sudo_debug.h>
+#include <sudo_eventlog.h>
+#include <sudo_gettext.h>
+#include <sudo_iolog.h>
+#include <sudo_fatal.h>
+#include <sudo_queue.h>
+#include <sudo_util.h>
 
-#include "logsrvd.h"
+#include <logsrvd.h>
 
-static inline bool
-has_numval(InfoMessage *info)
+static bool
+type_matches(InfoMessage *info, const char *source,
+    InfoMessage__ValueCase value_case)
 {
-    return info->value_case == INFO_MESSAGE__VALUE_NUMVAL;
-}
+    const void *val = info->u.strval;	/* same for strlistval */
+    debug_decl(type_matches, SUDO_DEBUG_UTIL);
 
-static inline bool
-has_strval(InfoMessage *info)
-{
-    return info->value_case == INFO_MESSAGE__VALUE_STRVAL;
-}
-
-static inline bool
-has_strlistval(InfoMessage *info)
-{
-    return info->value_case == INFO_MESSAGE__VALUE_STRLISTVAL;
+    if (info->key == NULL) {
+	sudo_warnx(U_("%s: protocol error: NULL key"), source);
+	debug_return_bool(false);
+    }
+    if (info->value_case != value_case) {
+	sudo_warnx(U_("%s: protocol error: wrong type for %s"),
+	    source, info->key);
+	debug_return_bool(false);
+    }
+    if (value_case != INFO_MESSAGE__VALUE_NUMVAL && val == NULL) {
+	sudo_warnx(U_("%s: protocol error: NULL value found in %s"),
+	    source, info->key);
+	debug_return_bool(false);
+    }
+    debug_return_bool(true);
 }
 
 /*
@@ -95,13 +106,27 @@ strlist_copy(InfoMessage__StringList *strlist)
 
 bad:
     if (dst != NULL) {
-	while (i--)
-	    free(dst[i]);
+	while (i)
+	    free(dst[--i]);
 	free(dst);
     }
     debug_return_ptr(NULL);
 }
 
+/*
+ * Free a NULL-terminated string vector.
+ */
+static void
+strvec_free(char *vec[])
+{
+    if (vec != NULL) {
+	char **vp;
+	for (vp = vec; *vp != NULL; vp++)
+	    free(*vp);
+	free(vec);
+    }
+}
+
 /*
  * Fill in eventlog details from an AcceptMessage
  * Caller is responsible for freeing strings in struct eventlog.
@@ -132,12 +157,15 @@ evlog_new(TimeSpec *submit_time, InfoMessage **info_msgs, size_t infolen,
     }
 
     /* Client/peer IP address. */
-    evlog->peeraddr = closure->ipaddr;
+    if ((evlog->peeraddr = strdup(closure->ipaddr)) == NULL) {
+	sudo_warnx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
+	goto bad;
+    }
 
     /* Submit time. */
     if (submit_time != NULL) {
-	evlog->submit_time.tv_sec = submit_time->tv_sec;
-	evlog->submit_time.tv_nsec = submit_time->tv_nsec;
+	evlog->event_time.tv_sec = (time_t)submit_time->tv_sec;
+	evlog->event_time.tv_nsec = (long)submit_time->tv_nsec;
     }
 
     /* Default values */
@@ -154,210 +182,202 @@ evlog_new(TimeSpec *submit_time, InfoMessage **info_msgs, size_t infolen,
 	switch (key[0]) {
 	case 'c':
 	    if (strcmp(key, "columns") == 0) {
-		if (!has_numval(info)) {
-		    sudo_warnx(U_("%s: protocol error: wrong type for %s"),
-			source, "columns");
-		} else if (info->u.numval <= 0 || info->u.numval > INT_MAX) {
-		    errno = ERANGE;
-		    sudo_warn(U_("%s: %s"), source, "columns");
-		} else {
-		    evlog->columns = info->u.numval;
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_NUMVAL)) {
+		    if (info->u.numval <= 0 || info->u.numval > INT_MAX) {
+			errno = ERANGE;
+			sudo_warn(U_("%s: %s"), source, "columns");
+		    } else {
+			evlog->columns = (int)info->u.numval;
+		    }
 		}
 		continue;
 	    }
 	    if (strcmp(key, "command") == 0) {
-		if (has_strval(info)) {
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_STRVAL)) {
+		    free(evlog->command);
 		    if ((evlog->command = strdup(info->u.strval)) == NULL) {
 			sudo_warnx(U_("%s: %s"), __func__,
 			    U_("unable to allocate memory"));
 			goto bad;
 		    }
-		} else {
-		    sudo_warnx(U_("%s: protocol error: wrong type for %s"),
-			source, "command");
 		}
 		continue;
 	    }
 	    break;
 	case 'l':
 	    if (strcmp(key, "lines") == 0) {
-		if (!has_numval(info)) {
-		    sudo_warnx(U_("%s: protocol error: wrong type for %s"),
-			source, "lines");
-		} else if (info->u.numval <= 0 || info->u.numval > INT_MAX) {
-		    errno = ERANGE;
-		    sudo_warn(U_("%s: %s"), source, "lines");
-		} else {
-		    evlog->lines = info->u.numval;
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_NUMVAL)) {
+		    if (info->u.numval <= 0 || info->u.numval > INT_MAX) {
+			errno = ERANGE;
+			sudo_warn(U_("%s: %s"), source, "lines");
+		    } else {
+			evlog->lines = (int)info->u.numval;
+		    }
 		}
 		continue;
 	    }
 	    break;
 	case 'r':
 	    if (strcmp(key, "runargv") == 0) {
-		if (has_strlistval(info)) {
-		    evlog->argv = strlist_copy(info->u.strlistval);
-		    if (evlog->argv == NULL)
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_STRLISTVAL)) {
+		    strvec_free(evlog->runargv);
+		    evlog->runargv = strlist_copy(info->u.strlistval);
+		    if (evlog->runargv == NULL)
 			goto bad;
-		} else {
-		    sudo_warnx(U_("%s: protocol error: wrong type for %s"),
-			source, "runargv");
 		}
 		continue;
 	    }
 	    if (strcmp(key, "runchroot") == 0) {
-		if (has_strval(info)) {
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_STRVAL)) {
+		    free(evlog->runchroot);
 		    if ((evlog->runchroot = strdup(info->u.strval)) == NULL) {
 			sudo_warnx(U_("%s: %s"), __func__,
 			    U_("unable to allocate memory"));
 			goto bad;
 		    }
-		} else {
-		    sudo_warnx(U_("%s: protocol error: wrong type for %s"),
-			source, "runchroot");
 		}
 		continue;
 	    }
 	    if (strcmp(key, "runcwd") == 0) {
-		if (has_strval(info)) {
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_STRVAL)) {
+		    free(evlog->runcwd);
 		    if ((evlog->runcwd = strdup(info->u.strval)) == NULL) {
 			sudo_warnx(U_("%s: %s"), __func__,
 			    U_("unable to allocate memory"));
 			goto bad;
 		    }
-		} else {
-		    sudo_warnx(U_("%s: protocol error: wrong type for %s"),
-			source, "runcwd");
 		}
 		continue;
 	    }
 	    if (strcmp(key, "runenv") == 0) {
-		if (has_strlistval(info)) {
-		    evlog->envp = strlist_copy(info->u.strlistval);
-		    if (evlog->envp == NULL)
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_STRLISTVAL)) {
+		    strvec_free(evlog->runenv);
+		    evlog->runenv = strlist_copy(info->u.strlistval);
+		    if (evlog->runenv == NULL)
 			goto bad;
-		} else {
-		    sudo_warnx(U_("%s: protocol error: wrong type for %s"),
-			source, "runenv");
 		}
 		continue;
 	    }
 	    if (strcmp(key, "rungid") == 0) {
-		if (!has_numval(info)) {
-		    sudo_warnx(U_("%s: protocol error: wrong type for %s"),
-			source, "rungid");
-		} else if (info->u.numval < 0 || info->u.numval > INT_MAX) {
-		    errno = ERANGE;
-		    sudo_warn(U_("%s: %s"), source, "rungid");
-		} else {
-		    evlog->rungid = info->u.numval;
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_NUMVAL)) {
+		    if (info->u.numval < 0 || info->u.numval > UINT_MAX) {
+			errno = ERANGE;
+			sudo_warn(U_("%s: %s"), source, "rungid");
+		    } else {
+			evlog->rungid = (gid_t)info->u.numval;
+		    }
 		}
 		continue;
 	    }
 	    if (strcmp(key, "rungroup") == 0) {
-		if (has_strval(info)) {
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_STRVAL)) {
+		    free(evlog->rungroup);
 		    if ((evlog->rungroup = strdup(info->u.strval)) == NULL) {
 			sudo_warnx(U_("%s: %s"), __func__,
 			    U_("unable to allocate memory"));
 			goto bad;
 		    }
-		} else {
-		    sudo_warnx(U_("%s: protocol error: wrong type for %s"),
-			source, "rungroup");
 		}
 		continue;
 	    }
 	    if (strcmp(key, "runuid") == 0) {
-		if (!has_numval(info)) {
-		    sudo_warnx(U_("%s: protocol error: wrong type for %s"),
-			source, "runuid");
-		} else if (info->u.numval < 0 || info->u.numval > INT_MAX) {
-		    errno = ERANGE;
-		    sudo_warn(U_("%s: %s"), source, "runuid");
-		} else {
-		    evlog->runuid = info->u.numval;
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_NUMVAL)) {
+		    if (info->u.numval < 0 || info->u.numval > UINT_MAX) {
+			errno = ERANGE;
+			sudo_warn(U_("%s: %s"), source, "runuid");
+		    } else {
+			evlog->runuid = (uid_t)info->u.numval;
+		    }
 		}
 		continue;
 	    }
 	    if (strcmp(key, "runuser") == 0) {
-		if (has_strval(info)) {
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_STRVAL)) {
+		    free(evlog->runuser);
 		    if ((evlog->runuser = strdup(info->u.strval)) == NULL) {
 			sudo_warnx(U_("%s: %s"), __func__,
 			    U_("unable to allocate memory"));
 			goto bad;
 		    }
-		} else {
-		    sudo_warnx(U_("%s: protocol error: wrong type for %s"),
-			source, "runuser");
 		}
 		continue;
 	    }
 	    break;
 	case 's':
+	    if (strcmp(key, "source") == 0) {
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_STRVAL)) {
+		    free(evlog->source);
+		    if ((evlog->source = strdup(info->u.strval)) == NULL) {
+			sudo_warnx(U_("%s: %s"), __func__,
+			    U_("unable to allocate memory"));
+			goto bad;
+		    }
+		}
+		continue;
+	    }
 	    if (strcmp(key, "submitcwd") == 0) {
-		if (has_strval(info)) {
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_STRVAL)) {
+		    free(evlog->cwd);
 		    if ((evlog->cwd = strdup(info->u.strval)) == NULL) {
 			sudo_warnx(U_("%s: %s"), __func__,
 			    U_("unable to allocate memory"));
 			goto bad;
 		    }
-		} else {
-		    sudo_warnx(U_("%s: protocol error: wrong type for %s"),
-			source, "submitcwd");
+		}
+		continue;
+	    }
+	    if (strcmp(key, "submitenv") == 0) {
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_STRLISTVAL)) {
+		    strvec_free(evlog->submitenv);
+		    evlog->submitenv = strlist_copy(info->u.strlistval);
+		    if (evlog->submitenv == NULL)
+			goto bad;
 		}
 		continue;
 	    }
 	    if (strcmp(key, "submitgroup") == 0) {
-		if (has_strval(info)) {
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_STRVAL)) {
+		    free(evlog->submitgroup);
 		    if ((evlog->submitgroup = strdup(info->u.strval)) == NULL) {
 			sudo_warnx(U_("%s: %s"), __func__,
 			    U_("unable to allocate memory"));
 			goto bad;
 		    }
-		} else {
-		    sudo_warnx(U_("%s: protocol error: wrong type for %s"),
-			source, "submitgroup");
 		}
 		continue;
 	    }
 	    if (strcmp(key, "submithost") == 0) {
-		if (has_strval(info)) {
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_STRVAL)) {
+		    free(evlog->submithost);
 		    if ((evlog->submithost = strdup(info->u.strval)) == NULL) {
 			sudo_warnx(U_("%s: %s"), __func__,
 			    U_("unable to allocate memory"));
 			goto bad;
 		    }
-		} else {
-		    sudo_warnx(U_("%s: protocol error: wrong type for %s"),
-			source, "submithost");
 		}
 		continue;
 	    }
 	    if (strcmp(key, "submituser") == 0) {
-		if (has_strval(info)) {
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_STRVAL)) {
+		    free(evlog->submituser);
 		    if ((evlog->submituser = strdup(info->u.strval)) == NULL) {
 			sudo_warnx(U_("%s: %s"), __func__,
 			    U_("unable to allocate memory"));
 			goto bad;
 		    }
-		} else {
-		    sudo_warnx(U_("%s: protocol error: wrong type for %s"),
-			source, "submituser");
 		}
 		continue;
 	    }
 	    break;
 	case 't':
 	    if (strcmp(key, "ttyname") == 0) {
-		if (has_strval(info)) {
+		if (type_matches(info, source, INFO_MESSAGE__VALUE_STRVAL)) {
+		    free(evlog->ttyname);
 		    if ((evlog->ttyname = strdup(info->u.strval)) == NULL) {
 			sudo_warnx(U_("%s: %s"), __func__,
 			    U_("unable to allocate memory"));
 			goto bad;
 		    }
-		} else {
-		    sudo_warnx(U_("%s: protocol error: wrong type for %s"),
-			source, "ttyname");
 		}
 		continue;
 	    }
@@ -427,7 +447,7 @@ struct iolog_path_closure {
 };
 
 static size_t
-fill_seq(char *str, size_t strsize, void *v)
+fill_seq(char * restrict str, size_t strsize, void * restrict v)
 {
     struct iolog_path_closure *closure = v;
     char *sessid = closure->evlog->sessid;
@@ -446,11 +466,11 @@ fill_seq(char *str, size_t strsize, void *v)
 	sudo_warnx(U_("%s: unable to format session id"), __func__);
 	debug_return_size_t(strsize); /* handle non-standard snprintf() */
     }
-    debug_return_size_t(len);
+    debug_return_size_t((size_t)len);
 }
 
 static size_t
-fill_user(char *str, size_t strsize, void *v)
+fill_user(char * restrict str, size_t strsize, void * restrict v)
 {
     struct iolog_path_closure *closure = v;
     const struct eventlog *evlog = closure->evlog;
@@ -464,7 +484,7 @@ fill_user(char *str, size_t strsize, void *v)
 }
 
 static size_t
-fill_group(char *str, size_t strsize, void *v)
+fill_group(char * restrict str, size_t strsize, void * restrict v)
 {
     struct iolog_path_closure *closure = v;
     const struct eventlog *evlog = closure->evlog;
@@ -478,7 +498,7 @@ fill_group(char *str, size_t strsize, void *v)
 }
 
 static size_t
-fill_runas_user(char *str, size_t strsize, void *v)
+fill_runas_user(char * restrict str, size_t strsize, void * restrict v)
 {
     struct iolog_path_closure *closure = v;
     const struct eventlog *evlog = closure->evlog;
@@ -492,7 +512,7 @@ fill_runas_user(char *str, size_t strsize, void *v)
 }
 
 static size_t
-fill_runas_group(char *str, size_t strsize, void *v)
+fill_runas_group(char * restrict str, size_t strsize, void * restrict v)
 {
     struct iolog_path_closure *closure = v;
     const struct eventlog *evlog = closure->evlog;
@@ -507,7 +527,7 @@ fill_runas_group(char *str, size_t strsize, void *v)
 }
 
 static size_t
-fill_hostname(char *str, size_t strsize, void *v)
+fill_hostname(char * restrict str, size_t strsize, void * restrict v)
 {
     struct iolog_path_closure *closure = v;
     const struct eventlog *evlog = closure->evlog;
@@ -521,7 +541,7 @@ fill_hostname(char *str, size_t strsize, void *v)
 }
 
 static size_t
-fill_command(char *str, size_t strsize, void *v)
+fill_command(char * restrict str, size_t strsize, void * restrict v)
 {
     struct iolog_path_closure *closure = v;
     const struct eventlog *evlog = closure->evlog;
@@ -556,7 +576,7 @@ create_iolog_path(struct connection_closure *closure)
     struct eventlog *evlog = closure->evlog;
     struct iolog_path_closure path_closure;
     char expanded_dir[PATH_MAX], expanded_file[PATH_MAX], pathbuf[PATH_MAX];
-    size_t len;
+    int len;
     debug_decl(create_iolog_path, SUDO_DEBUG_UTIL);
 
     path_closure.evlog = evlog;
@@ -578,7 +598,7 @@ create_iolog_path(struct connection_closure *closure)
 
     len = snprintf(pathbuf, sizeof(pathbuf), "%s/%s", expanded_dir,
 	expanded_file);
-    if (len >= sizeof(pathbuf)) {
+    if (len < 0 || len >= ssizeof(pathbuf)) {
 	errno = ENAMETOOLONG;
 	sudo_warn("%s/%s", expanded_dir, expanded_file);
 	goto bad;
@@ -589,7 +609,7 @@ create_iolog_path(struct connection_closure *closure)
      * Calls mkdtemp() if pathbuf ends in XXXXXX.
      */
     if (!iolog_mkpath(pathbuf)) {
-	sudo_warnx(U_("unable to create iolog path %s"), pathbuf);
+	sudo_warn(U_("unable to create iolog path %s"), pathbuf);
         goto bad;
     }
     if ((evlog->iolog_path = strdup(pathbuf)) == NULL) {
@@ -632,14 +652,14 @@ void
 iolog_close_all(struct connection_closure *closure)
 {
     const char *errstr;
-    int i;
+    unsigned int i;
     debug_decl(iolog_close_all, SUDO_DEBUG_UTIL);
 
     for (i = 0; i < IOFD_MAX; i++) {
 	if (!closure->iolog_files[i].enabled)
 	    continue;
 	if (!iolog_close(&closure->iolog_files[i], &errstr)) {
-	    sudo_warnx(U_("error closing iofd %d: %s"), i, errstr);
+	    sudo_warnx(U_("error closing iofd %u: %s"), i, errstr);
 	}
     }
     if (closure->iolog_dir_fd != -1)
@@ -652,14 +672,15 @@ bool
 iolog_flush_all(struct connection_closure *closure)
 {
     const char *errstr;
-    int i, ret = true;
+    bool ret = true;
+    unsigned int i;
     debug_decl(iolog_flush_all, SUDO_DEBUG_UTIL);
 
     for (i = 0; i < IOFD_MAX; i++) {
 	if (!closure->iolog_files[i].enabled)
 	    continue;
 	if (!iolog_flush(&closure->iolog_files[i], &errstr)) {
-	    sudo_warnx(U_("error flushing iofd %d: %s"), i, errstr);
+	    sudo_warnx(U_("error flushing iofd %u: %s"), i, errstr);
 	    ret = false;
 	}
     }
@@ -709,14 +730,14 @@ iolog_copy(struct iolog_file *src, struct iolog_file *dst, off_t remainder,
     sudo_debug_printf(SUDO_DEBUG_INFO|SUDO_DEBUG_LINENO,
 	"copying %lld bytes", (long long)remainder);
     while (remainder > 0) {
-	const ssize_t toread = MIN(remainder, ssizeof(buf));
+	const size_t toread = MIN((size_t)remainder, sizeof(buf));
 	nread = iolog_read(src, buf, toread, errstr);
 	if (nread == -1)
 	    debug_return_bool(false);
 	remainder -= nread;
 
 	do {
-	    ssize_t nwritten = iolog_write(dst, buf, nread, errstr);
+	    ssize_t nwritten = iolog_write(dst, buf, (size_t)nread, errstr);
 	    if (nwritten == -1)
 		debug_return_bool(false);
 	    nread -= nwritten;
@@ -763,6 +784,9 @@ iolog_rewrite(const struct timespec *target, struct connection_closure *closure)
     bool ret = false;
     debug_decl(iolog_rewrite, SUDO_DEBUG_UTIL);
 
+    memset(&timing, 0, sizeof(timing));
+    timing.decimal = ".";
+
     /* Parse timing file until we reach the target point. */
     /* TODO: use iolog_seekto with a callback? */
     for (;;) {
@@ -778,7 +802,7 @@ iolog_rewrite(const struct timespec *target, struct connection_closure *closure)
 		    evlog->iolog_path, iolog_fd_to_name(timing.event));
 		goto done;
 	    }
-	    iolog_file_sizes[timing.event] += timing.u.nbytes;
+	    iolog_file_sizes[timing.event] += (off_t)timing.u.nbytes;
 	}
 
 	if (sudo_timespeccmp(&closure->elapsed_time, target, >=)) {
@@ -903,12 +927,16 @@ update_elapsed_time(TimeSpec *delta, struct timespec *elapsed)
     debug_decl(update_elapsed_time, SUDO_DEBUG_UTIL);
 
     /* Cannot use timespecadd since msg doesn't use struct timespec. */
-    elapsed->tv_sec += delta->tv_sec;
-    elapsed->tv_nsec += delta->tv_nsec;
+    elapsed->tv_sec += (time_t)delta->tv_sec;
+    elapsed->tv_nsec += (long)delta->tv_nsec;
     while (elapsed->tv_nsec >= 1000000000) {
 	elapsed->tv_sec++;
 	elapsed->tv_nsec -= 1000000000;
     }
+    sudo_debug_printf(SUDO_DEBUG_DEBUG|SUDO_DEBUG_LINENO,
+	"%s: delta [%lld, %d], elapsed time now [%lld, %ld]",
+	__func__, (long long)delta->tv_sec, delta->tv_nsec,
+	(long long)elapsed->tv_sec, elapsed->tv_nsec);
 
     debug_return;
 }
