@@ -1,7 +1,7 @@
 /*
  * SPDX-License-Identifier: ISC
  *
- * Copyright (c) 2021 Todd C. Miller <Todd.Miller@sudo.ws>
+ * Copyright (c) 2021-2022 Todd C. Miller <Todd.Miller@sudo.ws>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -16,7 +16,12 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-#include "config.h"
+/*
+ * This is an open source non-commercial project. Dear PVS-Studio, please check it.
+ * PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com
+ */
+
+#include <config.h>
 
 #include <sys/stat.h>
 #include <sys/types.h>
@@ -31,7 +36,7 @@
 #ifdef HAVE_STDBOOL_H
 # include <stdbool.h>
 #else
-# include "compat/stdbool.h"
+# include <compat/stdbool.h>
 #endif /* HAVE_STDBOOL_H */
 #if defined(HAVE_STDINT_H)
 # include <stdint.h>
@@ -44,17 +49,17 @@
 #include <time.h>
 #include <unistd.h>
 
-#include "sudo_compat.h"
-#include "sudo_conf.h"
-#include "sudo_debug.h"
-#include "sudo_event.h"
-#include "sudo_eventlog.h"
-#include "sudo_fatal.h"
-#include "sudo_gettext.h"
-#include "sudo_iolog.h"
-#include "sudo_util.h"
+#include <sudo_compat.h>
+#include <sudo_conf.h>
+#include <sudo_debug.h>
+#include <sudo_event.h>
+#include <sudo_eventlog.h>
+#include <sudo_fatal.h>
+#include <sudo_gettext.h>
+#include <sudo_iolog.h>
+#include <sudo_util.h>
 
-#include "logsrvd.h"
+#include <logsrvd.h>
 
 /*
  * Helper function to set closure->journal and closure->journal_path.
@@ -65,6 +70,7 @@ journal_fdopen(int fd, const char *journal_path,
 {
     debug_decl(journal_fdopen, SUDO_DEBUG_UTIL);
 
+    free(closure->journal_path);
     closure->journal_path = strdup(journal_path);
     if (closure->journal_path == NULL) {
 	sudo_warnx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
@@ -72,6 +78,8 @@ journal_fdopen(int fd, const char *journal_path,
     }
 
     /* Defer fdopen() until last--it cannot be undone. */
+    if (closure->journal != NULL)
+	fclose(closure->journal);
     if ((closure->journal = fdopen(fd, "r+")) == NULL) {
 	sudo_debug_printf(SUDO_DEBUG_ERROR|SUDO_DEBUG_LINENO|SUDO_DEBUG_ERRNO,
 	    "unable to fdopen journal file %s", journal_path);
@@ -82,10 +90,11 @@ journal_fdopen(int fd, const char *journal_path,
 }
 
 static int
-journal_mkstemp(const char *parent_dir, char *pathbuf, int pathlen)
+journal_mkstemp(const char *parent_dir, char *pathbuf, size_t pathsize)
 {
-    int len, fd = -1;
+    int len, dfd = -1, fd = -1;
     mode_t dirmode, oldmask;
+    char *template;
     debug_decl(journal_mkstemp, SUDO_DEBUG_UTIL);
 
     /* umask must not be more restrictive than the file modes. */
@@ -96,27 +105,31 @@ journal_mkstemp(const char *parent_dir, char *pathbuf, int pathlen)
         dirmode |= S_IXOTH;
     oldmask = umask(ACCESSPERMS & ~dirmode);
 
-    len = snprintf(pathbuf, pathlen, "%s/%s/%s",
+    len = snprintf(pathbuf, pathsize, "%s/%s/%s",
 	logsrvd_conf_relay_dir(), parent_dir, RELAY_TEMPLATE);
-    if (len >= pathlen) {
+    if ((size_t)len >= pathsize) {
 	errno = ENAMETOOLONG;
 	sudo_warn("%s/%s/%s", logsrvd_conf_relay_dir(), parent_dir,
 	    RELAY_TEMPLATE);
 	goto done;
     }
-    if (!sudo_mkdir_parents(pathbuf, logsrvd_conf_iolog_uid(),
-	    logsrvd_conf_iolog_gid(), S_IRWXU|S_IXGRP|S_IXOTH, false)) {
+    dfd = sudo_open_parent_dir(pathbuf, logsrvd_conf_iolog_uid(),
+	logsrvd_conf_iolog_gid(), S_IRWXU|S_IXGRP|S_IXOTH, false);
+    if (dfd == -1) {
 	sudo_debug_printf(SUDO_DEBUG_ERROR|SUDO_DEBUG_LINENO|SUDO_DEBUG_ERRNO,
 	    "unable to create parent dir for %s", pathbuf);
 	goto done;
     }
-    if ((fd = mkstemp(pathbuf)) == -1) {
+    template = &pathbuf[(size_t)len - (sizeof(RELAY_TEMPLATE) - 1)];
+    if ((fd = mkostempsat(dfd, template, 0, 0)) == -1) {
 	sudo_warn(U_("%s: %s"), "mkstemp", pathbuf);
 	goto done;
     }
 
 done:
     umask(oldmask);
+    if (dfd != -1)
+	close(dfd);
 
     debug_return_int(fd);
 }
@@ -255,6 +268,12 @@ journal_seek(struct timespec *target, struct connection_closure *closure)
 
 	    if (msg_len > bufsize) {
 		bufsize = sudo_pow2_roundup(msg_len);
+		if (bufsize < msg_len) {
+		    /* overflow */
+		    errno = ENOMEM;
+		    closure->errstr = _("unable to allocate memory");
+		    break;
+		}
 		free(buf);
 		if ((buf = malloc(bufsize)) == NULL) {
 		    closure->errstr = _("unable to allocate memory");
@@ -280,7 +299,7 @@ journal_seek(struct timespec *target, struct connection_closure *closure)
 	client_message__free_unpacked(msg, NULL);
 	msg = client_message__unpack(NULL, msg_len, buf);
 	if (msg == NULL) {
-	    sudo_warnx("unable to unpack %s size %zu", "ClientMessage",
+	    sudo_warnx(U_("unable to unpack %s size %zu"), "ClientMessage",
 		(size_t)msg_len);
 	    closure->errstr = _("invalid journal file, unable to restart");
 	    break;
@@ -358,13 +377,8 @@ journal_seek(struct timespec *target, struct connection_closure *closure)
 		msg->type_case, "ClientMessage", closure->journal_path);
 	    break;
 	}
-	if (delay != NULL) {
+	if (delay != NULL)
 	    update_elapsed_time(delay, &closure->elapsed_time);
-	    sudo_debug_printf(SUDO_DEBUG_DEBUG|SUDO_DEBUG_LINENO,
-		"%s: elapsed time now [%lld, %ld]", closure->journal_path,
-		(long long)closure->elapsed_time.tv_sec,
-		closure->elapsed_time.tv_nsec);
-	}
 
 	if (sudo_timespeccmp(&closure->elapsed_time, target, >=)) {
 	    if (sudo_timespeccmp(&closure->elapsed_time, target, ==)) {
@@ -429,8 +443,8 @@ journal_restart(RestartMessage *msg, uint8_t *buf, size_t buflen,
     }
 
     /* Seek forward to resume point. */
-    target.tv_sec = msg->resume_point->tv_sec;
-    target.tv_nsec = msg->resume_point->tv_nsec;
+    target.tv_sec = (time_t)msg->resume_point->tv_sec;
+    target.tv_nsec = (long)msg->resume_point->tv_nsec;
     if (!journal_seek(&target, closure)) {
 	sudo_warn(U_("unable to seek to [%lld, %ld] in journal file %s"),
 	    (long long)target.tv_sec, target.tv_nsec, journal_path);
@@ -441,7 +455,7 @@ journal_restart(RestartMessage *msg, uint8_t *buf, size_t buflen,
 }
 
 static bool
-journal_write(uint8_t *buf, size_t len, struct connection_closure *closure)
+journal_write(uint8_t * restrict buf, size_t len, struct connection_closure * restrict closure)
 {
     uint32_t msg_len;
     debug_decl(journal_write, SUDO_DEBUG_UTIL);
@@ -470,7 +484,7 @@ journal_accept(AcceptMessage *msg, uint8_t *buf, size_t len,
     debug_decl(journal_accept, SUDO_DEBUG_UTIL);
 
     if (closure->journal_path != NULL) {
-	/* Re-use existing journal file. */
+	/* Reuse existing journal file. */
 	debug_return_bool(journal_write(buf, len, closure));
     }
 
@@ -498,8 +512,8 @@ journal_accept(AcceptMessage *msg, uint8_t *buf, size_t len,
  * Store a RejectMessage from the client in the journal.
  */
 static bool
-journal_reject(RejectMessage *msg, uint8_t *buf, size_t len,
-    struct connection_closure *closure)
+journal_reject(RejectMessage *msg, uint8_t * restrict buf, size_t len,
+    struct connection_closure * restrict closure)
 {
     debug_decl(journal_reject, SUDO_DEBUG_UTIL);
 
@@ -518,8 +532,8 @@ journal_reject(RejectMessage *msg, uint8_t *buf, size_t len,
  * Store an ExitMessage from the client in the journal.
  */
 static bool
-journal_exit(ExitMessage *msg, uint8_t *buf, size_t len,
-    struct connection_closure *closure)
+journal_exit(ExitMessage *msg, uint8_t * restrict buf, size_t len,
+    struct connection_closure * restrict closure)
 {
     debug_decl(journal_exit, SUDO_DEBUG_UTIL);
 
@@ -536,20 +550,28 @@ journal_exit(ExitMessage *msg, uint8_t *buf, size_t len,
  * Store an AlertMessage from the client in the journal.
  */
 static bool
-journal_alert(AlertMessage *msg, uint8_t *buf, size_t len,
-    struct connection_closure *closure)
+journal_alert(AlertMessage *msg, uint8_t * restrict buf, size_t len,
+    struct connection_closure * restrict closure)
 {
     debug_decl(journal_alert, SUDO_DEBUG_UTIL);
 
-    debug_return_bool(journal_write(buf, len, closure));
+    /* Store message in a journal for later relaying. */
+    if (closure->journal_path == NULL) {
+	if (!journal_create(closure))
+	    debug_return_bool(false);
+    }
+    if (!journal_write(buf, len, closure))
+	debug_return_bool(false);
+
+    debug_return_bool(true);
 }
 
 /*
  * Store an IoBuffer from the client in the journal.
  */
 static bool
-journal_iobuf(int iofd, IoBuffer *iobuf, uint8_t *buf, size_t len,
-    struct connection_closure *closure)
+journal_iobuf(int iofd, IoBuffer *iobuf, uint8_t * restrict buf, size_t len,
+    struct connection_closure * restrict closure)
 {
     debug_decl(journal_iobuf, SUDO_DEBUG_UTIL);
 
@@ -564,11 +586,13 @@ journal_iobuf(int iofd, IoBuffer *iobuf, uint8_t *buf, size_t len,
  * Store a CommandSuspend message from the client in the journal.
  */
 static bool
-journal_suspend(CommandSuspend *msg, uint8_t *buf, size_t len,
-    struct connection_closure *closure)
+journal_suspend(CommandSuspend *msg, uint8_t * restrict buf, size_t len,
+    struct connection_closure * restrict closure)
 {
     debug_decl(journal_suspend, SUDO_DEBUG_UTIL);
 
+    update_elapsed_time(msg->delay, &closure->elapsed_time);
+
     debug_return_bool(journal_write(buf, len, closure));
 }
 
@@ -576,11 +600,13 @@ journal_suspend(CommandSuspend *msg, uint8_t *buf, size_t len,
  * Store a ChangeWindowSize message from the client in the journal.
  */
 static bool
-journal_winsize(ChangeWindowSize *msg, uint8_t *buf, size_t len,
-    struct connection_closure *closure)
+journal_winsize(ChangeWindowSize *msg, uint8_t * restrict buf, size_t len,
+    struct connection_closure * restrict closure)
 {
     debug_decl(journal_winsize, SUDO_DEBUG_UTIL);
 
+    update_elapsed_time(msg->delay, &closure->elapsed_time);
+
     debug_return_bool(journal_write(buf, len, closure));
 }
 
